Option Strict On
Option Explicit On
Imports System.Runtime.InteropServices

Public Class FormMain

    Const HEXA_LOG_MODE As Integer = 2

    Const FRAMEKEY_DATETIME As Integer = 96
    Const FRAMEKEY_RAIN As Integer = 65
    Const FRAMEKEY_TEMPERATURE As Integer = 66
    Const FRAMEKEY_WIND As Integer = 72
    Const FRAMEKEY_BAROMETER As Integer = 70

    Const FRAMELENGTH_DATETIME As Integer = 11
    Const FRAMELENGTH_RAIN As Integer = 16
    Const FRAMELENGTH_TEMPERATURE As Integer = 11
    Const FRAMELENGTH_WIND As Integer = 10
    Const FRAMELENGTH_BAROMETER As Integer = 7

    Const MSG_CI1 As String = "20 00 08 01 00 C7 F5 C8" 'CLASS INTERFACE
    Const MSG_CI2 As String = "07 A1 01 00 20 A1 01 00" 'CLASS INTERFACE
    Const MSG_CI3 As String = "01 20 01 00 20 A1 01 00" 'CLASS INTERFACE
    Const MSG_CT1 As String = "21 09 00 02 00 00 08 00" 'CONTROL TRANSFER

    'Const MSG_CT1 As String = "21 0A 00 00 00 00 00 00" 'CONTROL TRANSFER
    Const MSG_PCR As String = "01 D0 08 01 00 00 00 00"

    Dim nNotDetectedMessage As Integer = 0

    Private WithEvents oForm1 As Form1 = Nothing

    Dim oLastReception As New Collection
    
    Dim status As Integer = 0

    Dim bReading As Boolean = False

    Dim DeviceNotificationHandle As IntPtr
    Dim ExclusiveAccess As Boolean
    Dim HIDHandle As Integer
    Dim HIDUsage As String
    Dim MyDeviceDetected As Boolean
    Dim MyDevicePathName As String
    Dim MyHID As New Hid()
    Dim ReadHandle As Integer
    Dim WriteHandle As Integer

    Dim MyDeviceManagement As New DeviceManagement()

    Private Delegate Sub ReadInputReportDelegate _
        (ByVal readHandle As Integer, _
        ByVal hidHandle As Integer, _
        ByVal writeHandle As Integer, _
        ByRef myDeviceDetected As Boolean, _
        ByRef readBuffer() As Byte, _
        ByRef success As Boolean)

    Private Delegate Sub MarshalToForm _
        (ByVal action As String, _
        ByVal textToAdd As String)

    Private GlobalInputReportBuffer() As Byte
    Private GlobalInputReportBufferLength As Integer

    Private Sub GetVendorAndProductIDs(ByRef myVendorID As Short, ByRef myProductID As Short)

        'Purpose    : Retrieves a Vendor ID and Product ID in hexadecimal 
        '           : from the form's text boxes and converts the text to Shorts.

        'Accepts    : myVendorID - the Vendor ID as a Short.
        '           : myProductID - the Product ID as a Short.                    

        Try
            myVendorID = CShort(Val("&h" & TextBox_VendorID.Text))
            myProductID = CShort(Val("&h" & TextBox_ProductID.Text))

        Catch ex As Exception
        End Try
    End Sub

    Private Function FindTheHid() As Boolean

        'Purpose    : Uses a series of API calls to locate a HID-class device
        '           ; by its Vendor ID and Product ID.

        'Returns    : True if the device is detected, False if not detected.

        Dim DeviceFound As Boolean
        Dim DevicePathName(127) As String
        Dim GUIDString As String
        Dim HidGuid As System.Guid
        Dim LastDevice As Boolean
        Dim MemberIndex As Integer
        Dim MyProductID As Short
        Dim MyVendorID As Short
        Dim Result As Boolean
        Dim Security As SECURITY_ATTRIBUTES
        Dim Success As Boolean


        Try

            HidGuid = Guid.Empty
            LastDevice = False
            MyDeviceDetected = False

            'Values for the SECURITY_ATTRIBUTES structure:

            Security.lpSecurityDescriptor = 0
            Security.bInheritHandle = CInt(True)
            Security.nLength = Len(Security)

            'Get the device's Vendor ID and Product ID from the form's text boxes.

            GetVendorAndProductIDs(MyVendorID, MyProductID)

            '***
            'API function: 'HidD_GetHidGuid

            'Purpose: Retrieves the interface class GUID for the HID class.

            'Accepts: 'A System.Guid object for storing the GUID.
            '***

            HidD_GetHidGuid(HidGuid)

            'Display the GUID.

            GUIDString = HidGuid.ToString

            'Fill an array with the device path names of all attached HIDs.

            DeviceFound = MyDeviceManagement.FindDeviceFromGuid _
                (HidGuid, _
                DevicePathName)

            'If there is at least one HID, attempt to read the Vendor ID and Product ID
            'of each device until there is a match or all devices have been examined.

            If DeviceFound = True Then
                MemberIndex = 0
                Do
                    '***
                    'API function:
                    'CreateFile

                    'Purpose:
                    'Retrieves a handle to a device.

                    'Accepts:
                    'A device path name returned by SetupDiGetDeviceInterfaceDetail
                    'The type of access requested (read/write).
                    'FILE_SHARE attributes to allow other processes to access the device while this handle is open.
                    'A Security structure. Using Null for this may cause problems under Windows XP.
                    'A creation disposition value. Use OPEN_EXISTING for devices.
                    'Flags and attributes for files. Not used for devices.
                    'Handle to a template file. Not used.

                    'Returns: a handle without read or write access.
                    'This enables obtaining information about all HIDs, even system
                    'keyboards and mice. 
                    'Separate handles are used for reading and writing.
                    '***

                    HIDHandle = CreateFile _
                        (DevicePathName(MemberIndex), _
                        0, _
                        FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                        Security, _
                        OPEN_EXISTING, _
                        0, _
                        0)

                    If (HIDHandle <> INVALID_HANDLE_VALUE) Then

                        'The returned handle is valid, 
                        'so find out if this is the device we're looking for.

                        'Set the Size property of DeviceAttributes to the number of bytes in the structure.

                        MyHID.DeviceAttributes.Size = Marshal.SizeOf(MyHID.DeviceAttributes)

                        '***
                        'API function:
                        'HidD_GetAttributes

                        'Purpose:
                        'Retrieves a HIDD_ATTRIBUTES structure containing the Vendor ID, 
                        'Product ID, and Product Version Number for a device.

                        'Accepts:
                        'A handle returned by CreateFile.
                        'A pointer to receive a HIDD_ATTRIBUTES structure.

                        'Returns:
                        'True on success, False on failure.
                        '***

                        Result = HidD_GetAttributes(HIDHandle, MyHID.DeviceAttributes)


                        If Result Then
                            'Find out if the device matches the one we're looking for.

                            If (MyHID.DeviceAttributes.VendorID = MyVendorID) And _
                                (MyHID.DeviceAttributes.ProductID = MyProductID) Then

                                'It's the desired device.

                                'Display the information in form's list box.

                                MyDeviceDetected = True

                                'Save the DevicePathName so OnDeviceChange() knows which name is my device.

                                MyDevicePathName = DevicePathName(MemberIndex)
                            Else

                                'It's not a match, so close the handle.

                                MyDeviceDetected = False

                                Result = CloseHandle(HIDHandle)
                            End If
                        Else
                            'There was a problem in retrieving the information.

                            MyDeviceDetected = False
                            Result = CloseHandle(HIDHandle)
                        End If

                    End If


                    'Keep looking until we find the device or there are no more left to examine.

                    MemberIndex = MemberIndex + 1

                Loop Until ((MyDeviceDetected = True) Or _
                    (MemberIndex = UBound(DevicePathName) + 1))
            End If

            If MyDeviceDetected Then

                'The device was detected.
                'Register to receive notifications if the device is removed or attached.

                Success = MyDeviceManagement.RegisterForDeviceNotifications _
                    (MyDevicePathName, _
                    Me.Handle, _
                    HidGuid, _
                    DeviceNotificationHandle)

                Debug.WriteLine("RegisterForDeviceNotifications = " & Success)

                'Learn the capabilities of the device.

                MyHID.Capabilities = MyHID.GetDeviceCapabilities _
                    (HIDHandle)

                If Success Then

                    'Find out if the device is a system mouse or keyboard.

                    HIDUsage = MyHID.GetHIDUsage(MyHID.Capabilities)

                    'Get and display the Input report buffer size.

                    'Get handles to use in requesting Input and Output reports.

                    ReadHandle = CreateFile _
                        (MyDevicePathName, _
                        GENERIC_READ, _
                        FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                        Security, _
                        OPEN_EXISTING, _
                        FILE_FLAG_OVERLAPPED, _
                        0)
                    'ReadHandle = CreateFile _
                    '    (MyDevicePathName, _
                    '    GENERIC_READ, _
                    '    FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                    '    Security, _
                    '    OPEN_EXISTING, _
                    '    0, _
                    '    0)

                    If (ReadHandle = INVALID_HANDLE_VALUE) Then

                        ExclusiveAccess = True


                    Else

                        'WriteHandle = CreateFile _
                        '     (MyDevicePathName, _
                        '     GENERIC_WRITE, _
                        '     FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                        '     Security, _
                        '     OPEN_EXISTING, _
                        '     0, _
                        '     0)
                        WriteHandle = CreateFile _
                             (MyDevicePathName, _
                             GENERIC_WRITE, _
                             FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                             Security, _
                             OPEN_EXISTING, _
                             FILE_FLAG_OVERLAPPED, _
                             0)

                        '(optional)
                        'Flush any waiting reports in the input buffer.

                        'MyHID.FlushQueue(ReadHandle)

                    End If

                End If

                nNotDetectedMessage = 0
                MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : Device has been detected" & vbCrLf)
                MyMarshalToForm("Log2", vbCrLf)


            Else

                'The device wasn't detected.
                MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : Device not detected" & vbCrLf)
                MyMarshalToForm("Log2", vbCrLf)
                nNotDetectedMessage += 1
            End If
            Return MyDeviceDetected

        Catch ex As Exception
        End Try
    End Function

    Private Sub Shutdown()

        'Purpose    : Perform actions that must execute when the program ends.

        Try
            'Close open handles to the device.

            If (HIDHandle <> 0) Then
                CloseHandle(HIDHandle)
            End If

            If (ReadHandle <> 0) Then
                CloseHandle(ReadHandle)
            End If

            If (WriteHandle <> 0) Then
                CloseHandle(WriteHandle)
            End If

            'Stop receiving notifications.

            Call MyDeviceManagement.StopReceivingDeviceNotifications(DeviceNotificationHandle)

        Catch ex As Exception
        End Try

    End Sub


    Private Sub GetInputReportData(ByVal ar As IAsyncResult)

        'Purpose    : Retrieves Input report data and status information.
        '           : This routine is called automatically when myInputReport.Read
        '           : returns.
        '           : Calls several marshaling routines to access the main form.

        'Accepts    : ar - an object containing status information about 
        '           : the asynchronous operation.    


        Dim InputReportBuffer As Byte()
        Dim Success As Boolean

        Try

            InputReportBuffer = Nothing

            'Define a delegate using the IAsyncResult object.

            Dim deleg As ReadInputReportDelegate = _
                DirectCast(ar.AsyncState, ReadInputReportDelegate)

            'Get the IAsyncResult object and the values of other paramaters that the
            'BeginInvoke method passed ByRef.

            deleg.EndInvoke(MyDeviceDetected, InputReportBuffer, Success, ar)

            If (Not MyDeviceDetected) Then
                nNotDetectedMessage = 0
                MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : Device has been lost" & vbCrLf)
                MyMarshalToForm("Log2", vbCrLf)
                Shutdown()

            End If
            
            'Display the received report data in the form's list box.

            If (ar.IsCompleted And Success) Then
                OnDataRead(InputReportBuffer)
            End If

        Catch ex As Exception
        End Try

    End Sub

    Private Function SendToHID(ByVal sMSG As String) As Boolean
        If (MyDeviceDetected = False) Then
            MyDeviceDetected = FindTheHid()
        End If

        Dim Success As Boolean
        Dim OutputReportBuffer() As Byte
        ReDim OutputReportBuffer(MyHID.Capabilities.OutputReportByteLength - 1)
        OutputReportBuffer(0) = 0

        Dim values As String() = sMSG.Split(" "c)
        Dim value As String
        Dim i As Integer = 1
        For Each value In values
            OutputReportBuffer(i) = Byte.Parse(value, Globalization.NumberStyles.HexNumber)
            i += 1
        Next

        'Use a control transfer to send the report,
        'even if the HID has an interrupt OUT endpoint.

        Dim myOutputReport As New Hid.OutputReport
        Success = myOutputReport.Write(OutputReportBuffer, WriteHandle)

        Console.WriteLine(Format(Now, "HH:mm:ss") & " : " & "Send " & sMSG)
        If HEXA_LOG_MODE = 1 Then
            MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : Sent " & sMSG & vbCrLf)
        End If

        'Application.DoEvents()

        Return Success
    End Function


    Private Sub Start(ByVal bRestart As Boolean)
        Dim InputReportBuffer() As Byte
        'Dim Success As Boolean

        If Not bRestart Then
            Timer1.Enabled = True
            TextBox1.Text = ""
            TextBox2.Text = ""
            TextBox1.SelectionStart = 0
            TextBox1.SelectionLength = 0
        End If

        If (MyDeviceDetected = False) Then
            MyDeviceDetected = FindTheHid()
        End If

        If MyDeviceDetected Then
            ReDim GlobalInputReportBuffer(1000)
            GlobalInputReportBufferLength = 0

            SendToHID(MSG_CI1)
            status = 2

            ReDim InputReportBuffer(MyHID.Capabilities.InputReportByteLength - 1)

            WaitRead(InputReportBuffer)
        End If
    End Sub

    Private Function WaitRead(ByRef InputReportBuffer() As Byte) As Integer
        If (MyDeviceDetected = False) Then
            MyDeviceDetected = FindTheHid()
        End If

        Dim Success As Boolean

        Dim ar As IAsyncResult
        Dim myInputReport As New Hid.InputReport

        Dim count As Integer
        For count = 0 To UBound(InputReportBuffer)
            InputReportBuffer(count) = 0
        Next count

        'Define a delegate for the Read method of myInputReport.

        Dim MyReadInputReportDelegate As _
            New ReadInputReportDelegate(AddressOf myInputReport.Read)

        'The BeginInvoke method calls myInputReport.Read to attempt to read a report.
        'The method has the same parameters as the Read function,
        'plus two additional parameters:
        'GetInputReportData is the callback procedure that executes when the Read function returns.
        'MyReadInputReportDelegate is the asynchronous delegate object.
        'The last parameter can optionally be an object passed to the callback.

        ar = MyReadInputReportDelegate.BeginInvoke _
            (ReadHandle, _
            HIDHandle, _
            WriteHandle, _
            MyDeviceDetected, _
            InputReportBuffer, _
            Success, _
            New AsyncCallback(AddressOf GetInputReportData), _
            MyReadInputReportDelegate)
    End Function

    Private Function GetHex(ByRef bytes() As Byte, Optional ByVal n As Integer = -1) As String
        Dim s As String = ""
        Dim i As Integer
        Dim strByte As String

        If n = -1 Then
            n = bytes.Length - 1
        End If

        For i = 0 To n
            strByte = Hex(bytes(i))
            If (Len(strByte) < 2) Then strByte = "0" & strByte
            If s <> "" Then s &= " "
            s &= strByte
        Next i

        Return s
    End Function

    Private Function OnDataRead(ByRef InputReportBuffer() As Byte) As Integer
        If bReading Then Exit Function

        bReading = True

        If status > 0 Then
            If status = 2 Then
                SendToHID(MSG_CT1)
                status = 3
            ElseIf status = 3 Then
                SendToHID(MSG_CI2)
                status = 4
            ElseIf status = 4 Then
                SendToHID(MSG_CT1)
                status = 5
            ElseIf status = 5 Then
                SendToHID(MSG_CI3)
                status = 6
            ElseIf status = 6 Then
                SendToHID(MSG_CT1)
                status = 0
            End If
        Else
            Console.WriteLine(Format(Now, "HH:mm:ss") & " : " & "(I) " & GetHex(InputReportBuffer))

            Dim i As Integer
            For i = 2 To 1 + InputReportBuffer(1) 'UBound(InputReportBuffer)
                GlobalInputReportBuffer(GlobalInputReportBufferLength + i - 2) = InputReportBuffer(i)
            Next i
            GlobalInputReportBufferLength += 1 + InputReportBuffer(1) - 1 'UBound(InputReportBuffer) - 1

            Console.WriteLine(Format(Now, "HH:mm:ss") & " : " & "(G) " & GetHex(GlobalInputReportBuffer, GlobalInputReportBufferLength - 1))

            For i = 0 To UBound(InputReportBuffer)
                InputReportBuffer(i) = 0
            Next i

            Dim count As Integer
            Dim byteValue As String
            Dim s As String = ""

            For count = GlobalInputReportBufferLength - 1 To 1 Step -1
                If GlobalInputReportBuffer(count) <> 0 Or s <> "" Then
                    'Add a leading zero to values from 0 to F.
                    If Len(Hex(GlobalInputReportBuffer(count))) < 2 Then
                        byteValue = "0" & Hex(GlobalInputReportBuffer(count))
                    Else
                        byteValue = Hex(GlobalInputReportBuffer(count))
                    End If

                    If s <> "" Then s = " " & s
                    s = byteValue & s
                End If
            Next count
            'lstResults.Items.Add(s)
            'Console.WriteLine(s)
            If HEXA_LOG_MODE = 1 Then
                MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : Received " & s & vbCrLf)
            End If

            Dim frameKeys As Integer() = {FRAMEKEY_DATETIME, _
                                          FRAMEKEY_RAIN, _
                                          FRAMEKEY_TEMPERATURE, _
                                          FRAMEKEY_WIND, _
                                          FRAMEKEY_BAROMETER}
            Dim frameLengths As Integer() = {FRAMELENGTH_DATETIME, _
                                          FRAMELENGTH_RAIN, _
                                          FRAMELENGTH_TEMPERATURE, _
                                          FRAMELENGTH_WIND, _
                                          FRAMELENGTH_BAROMETER}

            Dim iMaxFrame As Integer = 0

            Dim f As Integer
            Dim j As Integer
            Dim checkSum As Long
            For i = 0 To GlobalInputReportBufferLength - 1
                For f = 0 To frameKeys.Length - 1
                    If GlobalInputReportBuffer(i + 1) = frameKeys(f) Then
                        checkSum = 0
                        If (i + frameLengths(f) < GlobalInputReportBufferLength) Then
                            For j = i To i + frameLengths(f) - 2
                                checkSum += GlobalInputReportBuffer(j)
                            Next j
                            checkSum = checkSum
                            If (GlobalInputReportBuffer(i + frameLengths(f) - 1) + 256 * GlobalInputReportBuffer(i + frameLengths(f)) = checkSum) Then
                                'If HEXA_LOG_MODE = 2 Then
                                '    MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : Frame start in buffer : " & i & vbCrLf)
                                'End If
                                'Frame Found and CheckSum OK
                                Dim bytes(frameLengths(f)) As Byte
                                For j = i To i + frameLengths(f)
                                    bytes(j - i) = GlobalInputReportBuffer(j)
                                Next j
                                TranslateFrame(bytes)
                                iMaxFrame = i + frameLengths(f)
                            End If
                        End If
                    End If
                Next f
            Next i

            If iMaxFrame > 0 Then
                For i = iMaxFrame + 1 To GlobalInputReportBufferLength
                    GlobalInputReportBuffer(i - iMaxFrame - 1) = GlobalInputReportBuffer(i)
                Next i
                GlobalInputReportBufferLength -= iMaxFrame + 1
                'GlobalInputReportBufferLength = 0
                'For i = GlobalInputReportBufferLength To UBound(GlobalInputReportBuffer)
                'GlobalInputReportBuffer(i) = 0
                'Next i

                'MyHID.FlushQueue(ReadHandle)

                'SendToHID(MSG_CI2)
                'SendToHID(MSG_PCR)
                status = 0 '4
            End If
        End If

        System.Threading.Thread.Sleep(100)
        'Application.DoEvents()

        bReading = False

        WaitRead(InputReportBuffer)
        Return 0
    End Function

    Private Sub MyMarshalToForm _
        (ByVal action As String, _
        ByVal textToDisplay As String)

        'Purpose    : Enables accessing a form's controls from another thread 

        'Accepts    : action - a string that names the action to perform on the form
        '           : formText - text that the form displays or the code uses for 
        '           : another purpose. Actions that don't use text ignore this parameter.  

        Dim args() As Object = {action, textToDisplay}
        Dim MarshalToFormDelegate As MarshalToForm

        ' The AccessForm routine contains the code that accesses the form.

        MarshalToFormDelegate = New MarshalToForm(AddressOf AccessForm)

        ' Execute AccessForm, passing the parameters in args.

        MyBase.Invoke(MarshalToFormDelegate, args)

    End Sub

    Private Sub AccessForm(ByVal action As String, ByVal formText As String)

        'Purpose    : In asynchronous ReadFiles, the callback function GetInputReportData  
        '           : uses this routine to access the application's Form, which runs in 
        '           : a different thread.
        '           : The routine performs various application-specific functions that
        '           : involve accessing the application's form.

        'Accepts    : action - a string that names the action to perform on the form
        '           : formText - text that the form displays or the code uses for 
        '           : another purpose. Actions that don't use text ignore this parameter.  

        Try

            ' Select an action to perform on the form:

            Dim bDeletePreviousLine As Boolean = (nNotDetectedMessage >= 1)

            Dim oTextBox As TextBox = Nothing

            Select Case action
                Case "Log1"
                    oTextBox = TextBox1
                Case "Log2"
                    oTextBox = TextBox2
                Case Else
            End Select

            If Not IsNothing(oTextBox) Then
                If bDeletePreviousLine Then
                    Dim iPos As Integer = oTextBox.Text.LastIndexOf(vbCrLf, oTextBox.Text.Length - 2)
                    oTextBox.Text = oTextBox.Text.Remove(iPos + 2)
                End If
                oTextBox.Text = oTextBox.Text & formText
                oTextBox.SelectionStart = oTextBox.Text.Length
                oTextBox.ScrollToCaret()
            End If

        Catch ex As Exception
        End Try

    End Sub

    Private Function TranslateFrame(ByRef Bytes() As Byte) As Integer
        Dim s As String = ""
        Dim i As Integer

        'Dim strByte As String
        'For i = 0 To Bytes.Length - 1

        '    strByte = Hex(Bytes(i))
        '    If Len(strByte) < 2 Then
        '        strByte = "0" & strByte
        '    End If

        '    If s <> "" Then s &= " "
        '    s &= strByte
        'Next i
        s = GetHex(Bytes)

        If HEXA_LOG_MODE = 2 Then
            MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : " & s & vbCrLf)
        End If

        Dim checkSum As Integer
        For i = 0 To Bytes.Length - 3
            checkSum += Bytes(i)
        Next i
        'checkSum = checkSum Mod 256

        Dim bCS As Boolean = False

        Console.Write(Format(Now, "HH:mm:ss") & " : " & "[ " & s & " ]")
        'MyMarshalToForm("Log2", Format(Now, "HH:mm:ss") & " : [ " & s & " ]" & vbCrLf)
        If (checkSum = Bytes(Bytes.Length - 2) + 256 * Bytes(Bytes.Length - 1)) Then
            Console.WriteLine(" CheckSum OK")
            bCS = True
        Else
            Console.WriteLine()
        End If

        If bCS Then
            s = "OK : "
        Else
            s = "     : "
        End If

        Dim sTime As String = ""
        Dim lr As Date = Date.FromOADate(0)
        If bCS Then
            Dim sKey As String = Hex(Bytes(1))
            If Bytes(1) = FRAMEKEY_TEMPERATURE Then

                Dim iSensor As Integer = Bytes(2) Mod 16
                sKey &= iSensor
            End If
            Try
                lr = CDate(oLastReception(sKey))
            Catch ex As Exception
            End Try

            If (lr.ToOADate = 0) Then
                sTime = " "
            Else
                sTime = Int((Now.ToOADate - lr.ToOADate) * 24 * 60 * 60) & "s"
            End If

            Try
                oLastReception.Remove(sKey)
            Catch ex As Exception
            End Try
            oLastReception.Add(Now, sKey)
        End If

        s &= " " & sTime & " : "

        If (Bytes(1) = FRAMEKEY_WIND) Or (Bytes(1) = FRAMEKEY_DATETIME) Or (Bytes(1) = FRAMEKEY_RAIN) Then
            Dim iPower As Integer = Bytes(0) \ 16
            Dim bPower As Boolean = (iPower Mod 8) < 4
            Dim bWithoutSector As Boolean = iPower >= 8

            If bWithoutSector Then
                s &= "Secteur dbranch ; "
            End If
            If bPower Then
                s &= "Piles charges ; "
            Else
                s &= "Piles vides ; "
            End If
        End If

        'Anmomtre
        If (Bytes(1) = FRAMEKEY_WIND) Then
            Dim dWindGust As Double = CDbl(Bytes(4) + 256 * (Bytes(5) Mod 16)) * 0.1
            Dim dWindAverage As Double = CDbl(Bytes(5) \ 16 + Bytes(6) * 16) * 0.1
            Dim dWindDirection As Double = (CDbl(Bytes(2)) Mod 16) * 360 / 16


            s &= "Wind : "
            s &= "(G) " & dWindGust & " m/s ; "
            's &= "G " & CDbl(Bytes(4)) * 0.1 & " m/s ; "
            s &= "(M) " & dWindAverage & " m/s ; "
            s &= dWindDirection & ""

            MyMarshalToForm("Log2", Format(Now, "HH:mm:ss") & " : " & s & vbCrLf)


            If Not oForm1 Is Nothing Then
                oForm1.WeatherDataManager.Add(WeatherDataManager.C_WINDGUST, dWindGust)
                oForm1.WeatherDataManager.Add(WeatherDataManager.C_WINDSPEED, dWindAverage)
                oForm1.WeatherDataManager.Add(WeatherDataManager.C_WINDDIRECTION, dWindDirection)
            End If
        End If

        'Thermomtre/Hygromtre
        If Bytes(1) = FRAMEKEY_TEMPERATURE Then

            Dim iSensor As Integer = Bytes(2) Mod 16
            Dim dTemperature As Double = CDbl(Bytes(3) + Bytes(4) * 16 * 16) / 10
            Dim dHumidity As Double = CDbl(Bytes(5))
            Dim dDewPoint As Double = CDbl(Bytes(6) + Bytes(7) * 16 * 16) / 10
            Dim iConfort As Integer = Bytes(2) \ 16
            Dim iTempVariation As Integer = Bytes(0) \ 16

            s &= "Sensor " & iSensor & " : "
            s &= dTemperature & "C "
            Select Case iTempVariation
                Case 0 : s &= "(=) ; "
                Case 1 : s &= "(+) ; "
                Case 2 : s &= "(-) ; "
                Case Else : s &= "(" & iTempVariation & ") ; "
            End Select
            s &= dHumidity & "% ; "
            s &= dDewPoint & "C ; "
            s &= "Confort(" & iConfort & ") "
            Select Case iConfort
                Case 8 : s &= "Non convenable"
                Case 12 : s &= "Neutre"
                Case Else : s &= "?"
            End Select

            MyMarshalToForm("Log2", Format(Now, "HH:mm:ss") & " : " & s & vbCrLf)

            If Not oForm1 Is Nothing Then
                If iSensor = 0 Then
                    oForm1.WeatherDataManager.Add(WeatherDataManager.C_INSIDETEMPERATURE, dTemperature)
                    oForm1.WeatherDataManager.Add(WeatherDataManager.C_INSIDEHUMIDITY, dHumidity)
                End If
                If iSensor = 1 Then
                    oForm1.WeatherDataManager.Add(WeatherDataManager.C_OUTSIDETEMPERATURE, dTemperature)
                    oForm1.WeatherDataManager.Add(WeatherDataManager.C_OUTSIDEHUMIDITY, dHumidity)
                End If
            End If
        End If

        'Baromtre
        If (Bytes(1) = FRAMEKEY_BAROMETER) Then
            Dim dPressionRelative As Double = CDbl(Bytes(4) + (Bytes(5) Mod 16) * 16 * 16)
            Dim dPressionAbsolue As Double = CDbl(Bytes(2) + (Bytes(3) Mod 16) * 16 * 16)

            s &= "Baro : "
            s &= "PR=" & dPressionRelative & "mb ; "
            s &= "PA=" & dPressionAbsolue & "mb ; "

            Dim p As Integer

            p = Bytes(5) \ 16
            Select Case p
                Case 0 : s &= "Partly Cloudy"
                Case 1 : s &= "Rainy"
                Case 2 : s &= "Cloudy"
                Case 3 : s &= "Sunny"
                Case 4 : s &= "Snowy"
                Case Else : s &= "?"
            End Select

            s &= " => "

            p = Bytes(3) \ 16
            Select Case p
                Case 0 : s &= "Partly Cloudy"
                Case 1 : s &= "Rainy"
                Case 2 : s &= "Cloudy"
                Case 3 : s &= "Sunny"
                Case 4 : s &= "Snowy"
                Case Else : s &= "?"
            End Select

            MyMarshalToForm("Log2", Format(Now, "HH:mm:ss") & " : " & s & vbCrLf)

            If Not oForm1 Is Nothing Then
                oForm1.WeatherDataManager.Add(WeatherDataManager.C_BAROMETER, dPressionRelative)
            End If
        End If

        'Pluviomtre
        If (Bytes(1) = FRAMEKEY_RAIN) Then
            Dim dRainRate As Double = CDbl(Bytes(2) + Bytes(3) * 16 * 16) / 100 * 25.4
            Dim dRain1h As Double = CDbl(Bytes(4) + Bytes(5) * 16 * 16) / 100 * 25.4
            Dim dRain24h As Double = CDbl(Bytes(6) + Bytes(7) * 16 * 16) / 100 * 25.4
            Dim dRainAccum As Double = CDbl(Bytes(8) + Bytes(9) * 16 * 16) / 100 * 25.4

            s &= "Rain : "

            s &= "Rate = " & dRainRate & " mm/h ; "
            s &= "1h = " & dRain1h & " mm ; "
            s &= "24h  = " & dRain24h & " mm ; "
            s &= "Accum = " & dRainAccum & " mm "

            s &= "since " & Format(Bytes(12), "00") & "/" & Format(Bytes(13), "00") & "/" & Format(Bytes(14), "00") & " "
            s &= Format(Bytes(11), "00") & ":" & Format(Bytes(10), "00")

            MyMarshalToForm("Log2", Format(Now, "HH:mm:ss") & " : " & s & vbCrLf)

            If Not oForm1 Is Nothing Then
                oForm1.WeatherDataManager.Add(WeatherDataManager.C_RAIN, dRain1h)
            End If
        End If

        'Date/Heure
        If (Bytes(1) = FRAMEKEY_DATETIME) Then
            Dim iRadioFreq As Integer = (Bytes(0) \ 16) Mod 4

            s &= "Date/Time : "

            s &= Format(Bytes(6), "00") & "/" & Format(Bytes(7), "00") & "/" & Format(Bytes(8), "00") & " "
            s &= Format(Bytes(5), "00") & ":" & Format(Bytes(4), "00") & " ; "
            s &= "GMT "
            If Bytes(9) < 128 Then
                s &= "+"
            Else
                s &= "-"
            End If
            s &= Format(Bytes(9) Mod 128, "00") & " ; "
            Select Case iRadioFreq
                Case 0 : s &= " RadioFreq dsactive"
                Case 1 : s &= " RadioFreq en recherche ou signal faible"
                Case 2 : s &= " RadioFreq signal moyen"
                Case 3 : s &= " RadioFreq signal fort"
            End Select

            MyMarshalToForm("Log2", Format(Now, "HH:mm:ss") & " : " & s & vbCrLf)
        End If

    End Function

    'Private Sub FormMain_Activated(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Activated
    '    If Not IsNothing(oForm1) Then oForm1.BringToFront()
    '    'If Not Focused Then Focus()
    'End Sub

    Private Sub FormMain_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles Me.FormClosed
        If Not IsNothing(oForm1) Then oForm1.Close()
        Timer1.Enabled = False
        Shutdown()
    End Sub

    Protected Overrides Sub WndProc(ByRef m As Message)

        'Purpose    : Overrides WndProc to enable checking for and handling
        '           : WM_DEVICECHANGE(messages)

        'Accepts    : m - a Windows Message                   

        Try
            'The OnDeviceChange routine processes WM_DEVICECHANGE messages.

            If m.Msg = WM_DEVICECHANGE Then
                OnDeviceChange(m)
            End If

            'Let the base form process the message.

            MyBase.WndProc(m)

        Catch ex As Exception
        End Try

    End Sub

    Friend Sub OnDeviceChange(ByVal m As Message)

        'Purpose    : Called when a WM_DEVICECHANGE message has arrived,
        '           : indicating that a device has been attached or removed.

        'Accepts    : m - a message with information about the device

        Debug.WriteLine("WM_DEVICECHANGE")

        Try
            If (m.WParam.ToInt32 = DBT_DEVICEARRIVAL) Then

                'If WParam contains DBT_DEVICEARRIVAL, a device has been attached.

                'Find out if it's the device we're communicating with.

                If MyDeviceManagement.DeviceNameMatch(m, MyDevicePathName) Then
                    Start(True)
                End If

            ElseIf (m.WParam.ToInt32 = DBT_DEVICEREMOVECOMPLETE) Then

                'If WParam contains DBT_DEVICEREMOVAL, a device has been removed.

                Debug.WriteLine("A device has been removed.")

                'Find out if it's the device we're communicating with.

                If MyDeviceManagement.DeviceNameMatch(m, MyDevicePathName) Then

                    nNotDetectedMessage = 0
                    MyMarshalToForm("Log1", Format(Now, "HH:mm:ss") & " : Device has been removed" & vbCrLf)
                    MyMarshalToForm("Log2", vbCrLf)

                    Shutdown()

                    'Set MyDeviceDetected False so on the next data-transfer attempt,
                    'FindTheHid() will be called to look for the device 
                    'and get a new handle.

                    MyDeviceDetected = False
                End If
            End If
        Catch ex As Exception
        End Try
    End Sub

    Private Sub Timer1_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles Timer1.Tick
        If Not MyDeviceDetected Then Start(True)
    End Sub

    Private Sub FormMain_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        'oForm1 = New Form1
        'oForm1.Show()
        'FormMain_Move(sender, e)
        Start(False)
    End Sub

    'Private Sub oForm1_Activated(ByVal sender As Object, ByVal e As System.EventArgs) Handles oForm1.Activated
    '    Me.BringToFront()
    '    'If Not IsNothing(oForm1) AndAlso Not oForm1.Focused Then oForm1.Focus()
    'End Sub

    Private Sub oForm1_FormClosed(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosedEventArgs) Handles oForm1.FormClosed
        oForm1 = Nothing
        Close()
    End Sub

    Private Sub FormMain_Move(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Move
        If Not IsNothing(oForm1) Then
            oForm1.Left = Me.Left + Me.Width
            oForm1.Top = Me.Top
        End If
    End Sub

    Private Sub FormMain_SizeChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.SizeChanged
        If Not IsNothing(oForm1) Then
            oForm1.Left = Me.Left + Me.Width
            oForm1.Top = Me.Top
        End If
    End Sub
End Class